from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Iterator

import numpy as np
from viser import _messages

from . import _viser_private as impl
from ._audio import AudioHandle, AudioState, audio_array_payload
from ._protocol import AddAudioOp, AudioOp
from ._timeline import (
    TimelineRecorder,
    TimelineStore,
    is_scene_message,
    serialize_message,
)

if TYPE_CHECKING:
    from ._server import Viser4dServer


class SceneRecorder:
    def __init__(self, server: Viser4dServer, timeline: TimelineStore) -> None:
        self._server = server
        self._timeline = timeline
        self._active_step: int | None = None

    @property
    def active_step(self) -> int | None:
        return self._active_step

    @contextlib.contextmanager
    def at(self, t: int) -> Iterator[None]:
        step = self._timeline.validate_step(int(t))
        recorder = TimelineRecorder()
        self._active_step = step
        try:
            with impl.scene_recording_interface(self._server.scene, recorder):
                yield
        finally:
            self._active_step = None

        if not recorder.messages:
            return

        step_store = self._timeline.record_scene_messages(step, recorder.messages)
        for node_name in step_store.node_names:
            self._register_timeline_node(node_name)

        messages = [serialize_message(message) for message in recorder.messages]
        self._server._send_runtime_call(
            "preloadSceneStep",
            {
                "step": step,
                "messages": messages,
                "nodeNames": sorted(step_store.node_names),
            },
        )

    def add_audio(
        self,
        name: str,
        *,
        data: np.ndarray,
        sample_rate: int,
    ) -> AudioHandle:
        state = AudioState(
            name=name,
            sample_rate=int(sample_rate),
            waveform=np.ascontiguousarray(data),
        )
        handle = AudioHandle(self._server, state)
        assert self._active_step is not None
        op = AddAudioOp(
            op="add",
            name=name,
            sampleRate=state.sample_rate,
            waveform=audio_array_payload(state.waveform),
            volume=state.volume,
        )
        self._timeline.record_audio_ops(self._active_step, [op])
        self._server._send_runtime_call(
            "preloadAudioStep",
            {"step": self._active_step, "ops": [op]},
        )
        return handle

    def dispatch_audio_update(self, op: AudioOp) -> None:
        if self._active_step is not None:
            self._timeline.record_audio_ops(self._active_step, [op])
            self._server._send_runtime_call(
                "preloadAudioStep",
                {"step": self._active_step, "ops": [op]},
            )
            return
        self._server._send_runtime_call("applyAudioUpdate", op)

    def _register_timeline_node(self, name: str) -> None:
        if name in self._timeline.baseline_messages_by_name:
            return
        baseline = self._collect_live_messages_for_name(name)
        if not baseline:
            return
        self._timeline.baseline_messages_by_name[name] = baseline
        messages = [serialize_message(message) for message in baseline]
        self._server._send_runtime_call(
            "setBaseline",
            {"name": name, "messages": messages},
        )

    def _collect_live_messages_for_name(self, name: str) -> list[_messages.Message]:
        return [
            message
            for message in impl.broadcast_messages(self._server)
            if is_scene_message(message) and getattr(message, "name", None) == name
        ]



from __future__ import annotations

import base64
import pathlib

import msgspec
import numpy as np
import pytest
import zstandard

import viser4d
from viser4d._audio import audio_array_payload


def test_audio_requires_timestep_context() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with pytest.raises(RuntimeError):
            server.scene.add_audio(
                "/audio", data=np.zeros(4, dtype=np.int16), sample_rate=8_000
            )
    finally:
        server.stop()


def test_num_steps_must_be_positive() -> None:
    with pytest.raises(ValueError, match="num_steps must be >= 1"):
        viser4d.Viser4dServer(num_steps=0, port=0, verbose=False)


def test_timeline_records_scene_and_audio(tmp_path: pathlib.Path) -> None:
    server = viser4d.Viser4dServer(num_steps=3, port=0, verbose=False)
    try:
        seen_timesteps: list[int] = []
        server.on_timestep_change(seen_timesteps.append)

        with server.at(0):
            frame = server.scene.add_frame("/frame")
            audio = server.scene.add_audio(
                "/audio",
                data=np.array([0, 1, 2], dtype=np.int16),
                sample_rate=16_000,
            )
        with server.at(1):
            frame.position = (1.0, 2.0, 3.0)
            audio.volume = 0.25

        assert server._timeline.step(0).messages
        assert server._timeline.step(0).audio_ops
        assert server._timeline.step(1).messages
        assert server._timeline.step(1).audio_ops

        server.set_fps(24.0)
        assert server._fps == 24.0

        server.play(server._fps)
        assert server._is_playing is True
        server.pause()
        assert server._is_playing is False

        server.seek(2)
        assert server._current_timestep == 2
        assert seen_timesteps[-1] == 2

        blob = server.serialize()
        assert blob
        out_path = tmp_path / "scene.viser4d"
        server.write_recording(out_path)
        assert out_path.read_bytes() == blob

        size = int.from_bytes(blob[:8], "little")
        decoded = zstandard.ZstdDecompressor().decompress(blob[8:], size)
        payload = msgspec.msgpack.decode(decoded)
        assert set(payload) == {"durationSeconds", "messages", "viserVersion"}
        assert payload["durationSeconds"] == pytest.approx(2 / 30.0)
        assert not any(
            message["type"] == "RunJavascriptMessage"
            for _, message in payload["messages"]
        )
        assert any(time > 0.0 for time, _ in payload["messages"])
    finally:
        server.stop()


def test_serialize_rejects_invalid_timestep_range(tmp_path: pathlib.Path) -> None:
    server = viser4d.Viser4dServer(num_steps=3, port=0, verbose=False)
    try:
        with pytest.raises(AssertionError):
            server.serialize(start_timestep=2, end_timestep=1)
    finally:
        server.stop()


def test_stop_shuts_down_predictor_thread() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    predictor_thread = server._controller._predictor_thread
    assert predictor_thread.is_alive()

    server.stop()

    assert predictor_thread.is_alive() is False


def test_audio_payload_normalizes_integer_formats() -> None:
    int16_payload = audio_array_payload(np.array([-32768, 0, 32767], dtype=np.int16))
    int16_values = np.frombuffer(
        base64.b64decode(int16_payload["data"]), dtype=np.float32
    )
    assert int16_payload["dtype"] == "float32"
    assert int16_payload["numChannels"] == 1
    assert int16_payload["numFrames"] == 3
    assert np.allclose(
        int16_values, np.array([-1.0, 0.0, 32767 / 32768], dtype=np.float32)
    )

    int32_payload = audio_array_payload(
        np.array([-2147483648, 0, 2147483647], dtype=np.int32)
    )
    int32_values = np.frombuffer(
        base64.b64decode(int32_payload["data"]), dtype=np.float32
    )
    assert int32_payload["dtype"] == "float32"
    assert int32_payload["numChannels"] == 1
    assert int32_payload["numFrames"] == 3
    assert np.allclose(
        int32_values,
        np.array([-1.0, 0.0, 2147483647 / 2147483648], dtype=np.float32),
    )


def test_audio_append_keeps_chunked_state_until_waveform_is_read() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0):
            audio = server.scene.add_audio(
                "/audio",
                data=np.array([1, 2], dtype=np.int16),
                sample_rate=16_000,
            )

        audio.append(np.array([3, 4], dtype=np.int16))
        audio.append(np.array([5, 6], dtype=np.int16))

        expected = np.array([1, 2, 3, 4, 5, 6], dtype=np.int16)
        assert np.array_equal(audio.waveform, expected)
        assert np.array_equal(audio._state.waveform, expected)
    finally:
        server.stop()


def test_audio_rejects_non_mono_or_stereo_shapes() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0):
            with pytest.raises(ValueError, match="mono or stereo"):
                server.scene.add_audio(
                    "/audio",
                    data=np.zeros((2, 2, 2), dtype=np.float32),
                    sample_rate=16_000,
                )
            with pytest.raises(ValueError, match="mono or stereo"):
                server.scene.add_audio(
                    "/audio-3ch",
                    data=np.zeros((4, 3), dtype=np.float32),
                    sample_rate=16_000,
                )
    finally:
        server.stop()


def test_stereo_audio_append_preserves_channel_layout() -> None:
    server = viser4d.Viser4dServer(num_steps=2, port=0, verbose=False)
    try:
        with server.at(0):
            audio = server.scene.add_audio(
                "/audio",
                data=np.array([[1, 10], [2, 20]], dtype=np.int16),
                sample_rate=16_000,
            )

        audio.append(np.array([[3, 30], [4, 40]], dtype=np.int16))

        expected = np.array([[1, 10], [2, 20], [3, 30], [4, 40]], dtype=np.int16)
        assert np.array_equal(audio.waveform, expected)
        assert np.array_equal(audio._state.waveform, expected)

        with pytest.raises(ValueError, match="channel count"):
            audio.append(np.array([5, 6], dtype=np.int16))
    finally:
        server.stop()



from __future__ import annotations

from collections.abc import Mapping
from typing import Literal, TypeAlias, TypedDict


JSONValue: TypeAlias = (
    None | bool | int | float | str | list["JSONValue"] | dict[str, "JSONValue"]
)
SerializedMessage: TypeAlias = dict[str, JSONValue]


class BinaryPayload(TypedDict):
    __viser4d_binary__: str


class AudioArrayPayload(TypedDict):
    dtype: str
    numChannels: int
    numFrames: int
    data: str


class AddAudioOp(TypedDict):
    op: Literal["add"]
    name: str
    sampleRate: int
    waveform: AudioArrayPayload
    volume: float


class SetAudioVolumeOp(TypedDict):
    op: Literal["set_volume"]
    name: str
    volume: float


class SetAudioWaveformOp(TypedDict):
    op: Literal["set_waveform"]
    name: str
    waveform: AudioArrayPayload


class AppendAudioOp(TypedDict):
    op: Literal["append"]
    name: str
    waveform: AudioArrayPayload


class RemoveAudioOp(TypedDict):
    op: Literal["remove"]
    name: str


AudioOp: TypeAlias = (
    AddAudioOp | SetAudioVolumeOp | SetAudioWaveformOp | AppendAudioOp | RemoveAudioOp
)


class RuntimeConfig(TypedDict):
    numSteps: int
    fps: float
    baseFps: float
    loop: bool


class ClientRuntimeConfig(RuntimeConfig):
    timestepSyncUuid: str


RuntimeMethod: TypeAlias = Literal[
    "applyAudioUpdate",
    "configure",
    "pause",
    "play",
    "preloadAudioStep",
    "preloadSceneStep",
    "seek",
    "setBaseline",
    "setFps",
]

RuntimePayload: TypeAlias = AudioOp | Mapping[str, object]



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Iterator

import viser
from viser import _messages


def gui_uuid(handle: Any) -> str:
    return handle._impl.uuid


@contextmanager
def scene_recording_interface(
    scene: viser.SceneApi,
    recorder: Any,
) -> Iterator[None]:
    original_interface = scene._websock_interface
    scene._websock_interface = recorder
    try:
        yield
    finally:
        scene._websock_interface = original_interface


def broadcast_messages(server: viser.ViserServer) -> list[_messages.Message]:
    return [
        message
        for message in server._websock_server._broadcast_buffer.message_from_id.values()
        if isinstance(message, _messages.Message)
    ]



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from . import _viser_private as impl
from ._protocol import SerializedMessage
from ._runtime import RUNTIME_MARKER
from ._timeline import (
    TimelineStore,
    serialize_message,
    serialize_viser_recording,
)

if TYPE_CHECKING:
    from ._server import Viser4dServer


class ExportBuilder:
    def __init__(self, server: Viser4dServer, timeline: TimelineStore) -> None:
        self._server = server
        self._timeline = timeline

    def serialize(
        self, *, start_timestep: int = 0, end_timestep: int | None = None
    ) -> bytes:
        start = int(start_timestep)
        end = (
            int(end_timestep)
            if end_timestep is not None
            else self._server.num_steps - 1
        )
        assert 0 <= start < self._server.num_steps
        assert 0 <= end < self._server.num_steps
        assert start <= end

        recording: list[tuple[float, SerializedMessage]] = []
        for message in impl.broadcast_messages(self._server):
            if getattr(message, "source", "").startswith(RUNTIME_MARKER):
                continue
            if getattr(message, "name", None) in self._timeline.node_names:
                continue
            recording.append((0.0, serialize_message(message)))
        for baseline in self._timeline.baseline_messages_by_name.values():
            recording.extend((0.0, serialize_message(message)) for message in baseline)
        fps = max(self._server._base_fps, 1.0)
        for step in range(end + 1):
            time = 0.0 if step <= start else (step - start) / fps
            recording.extend(
                (time, serialize_message(message))
                for message in self._timeline.step(step).messages
            )

        blob = serialize_viser_recording(
            recording,
            duration_seconds=max(end - start, 0) / fps,
        )
        return blob

    def write(
        self,
        path: str | pathlib.Path,
        *,
        start_timestep: int = 0,
        end_timestep: int | None = None,
    ) -> None:
        blob = self.serialize(start_timestep=start_timestep, end_timestep=end_timestep)
        pathlib.Path(path).write_bytes(blob)



from __future__ import annotations

import base64
from dataclasses import dataclass, field
from typing import Any, cast

import msgspec
import numpy as np
import viser
import zstandard
from viser import _messages
from viser.infra import WebsockMessageHandler

from ._protocol import AudioOp, BinaryPayload, JSONValue, SerializedMessage


def to_jsonable(value: Any) -> JSONValue:
    if isinstance(value, memoryview):
        value = value.tobytes()
    if isinstance(value, (bytes, bytearray)):
        return cast(
            JSONValue,
            BinaryPayload(
                __viser4d_binary__=base64.b64encode(bytes(value)).decode("ascii")
            ),
        )
    if isinstance(value, np.ndarray):
        return cast(
            JSONValue,
            BinaryPayload(
                __viser4d_binary__=base64.b64encode(value.tobytes()).decode("ascii")
            ),
        )
    if isinstance(value, dict):
        return {str(key): to_jsonable(val) for key, val in value.items()}
    if isinstance(value, (tuple, list)):
        return [to_jsonable(item) for item in value]
    return value


def serialize_message(message: _messages.Message) -> SerializedMessage:
    return cast(SerializedMessage, to_jsonable(message.as_serializable_dict()))


def extract_node_names(message: _messages.Message) -> set[str]:
    name = getattr(message, "name", None)
    return {name} if isinstance(name, str) and name else set()


def is_scene_message(message: Any) -> bool:
    return "Gui" not in type(message).__name__


@dataclass
class TimelineStep:
    messages: list[_messages.Message] = field(default_factory=list)
    node_names: set[str] = field(default_factory=set)
    audio_ops: list[AudioOp] = field(default_factory=list)


class TimelineStore:
    def __init__(self, num_steps: int) -> None:
        self.num_steps = num_steps
        self.steps = [TimelineStep() for _ in range(num_steps)]
        self.node_names: set[str] = set()
        self.baseline_messages_by_name: dict[str, list[_messages.Message]] = {}

    def validate_step(self, step: int) -> int:
        if step < 0 or step >= self.num_steps:
            raise IndexError(
                f"Timestep {step} is out of range for {self.num_steps} steps."
            )
        return step

    def step(self, step: int) -> TimelineStep:
        return self.steps[self.validate_step(step)]

    def record_scene_messages(
        self, step: int, messages: list[_messages.Message]
    ) -> TimelineStep:
        step_state = self.step(step)
        step_state.messages.extend(messages)
        for message in messages:
            node_names = extract_node_names(message)
            step_state.node_names.update(node_names)
            self.node_names.update(node_names)
        return step_state

    def record_audio_ops(self, step: int, ops: list[AudioOp]) -> TimelineStep:
        step_state = self.step(step)
        step_state.audio_ops.extend(ops)
        return step_state


class TimelineRecorder(WebsockMessageHandler):
    def __init__(self) -> None:
        super().__init__()
        self.messages: list[_messages.Message] = []

    def get_message_buffer(self) -> Any:
        return self

    def push(self, message: _messages.Message) -> None:
        self.messages.append(message)

    def atomic_start(self) -> None:
        return

    def atomic_end(self) -> None:
        return


def serialize_viser_recording(
    messages: list[tuple[float, SerializedMessage]],
    *,
    duration_seconds: float = 0.0,
) -> bytes:
    packed = msgspec.msgpack.encode(
        {
            "durationSeconds": duration_seconds,
            "messages": messages,
            "viserVersion": viser.__version__,
        }
    )
    compressed = zstandard.ZstdCompressor(level=12).compress(packed)
    return len(packed).to_bytes(8, "little") + compressed
